package net.sourceforge.chopchophttpclient;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.net.ProxySelector;
import java.net.URI;
import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;

import net.sourceforge.chopchophttpclient.processor.DownloadRunnable;
import net.sourceforge.chopchophttpclient.processor.ProcessingThread;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.MultiThreadedHttpConnectionManager;
import org.apache.commons.httpclient.UsernamePasswordCredentials;
import org.apache.commons.httpclient.auth.AuthScope;
import org.apache.log4j.Logger;


/**
 * You <b>SHOULD</b> configure logging for the package org.apache.commons.httpclient otherwise your performance
 * will be very poor.
 * 
 * <p>Example:</p>
 * Logger.getLogger("org.apache.commons.httpclient").setLevel(Level.DEBUG);<br />
 * Logger.getLogger("httpclient.wire").setLevel(Level.OFF);
 *
 */
public class Client {
	private static final Logger log = Logger.getLogger(Client.class);
	
	
	private final String username;
	
	private final String password;
	
	private final ProxySelector proxySelector;
	
	private final ProgressListener progressListener;
	
	
	private final HttpClient httpClient;
	
	public static final int TIMEOUT_MS = 15 * 1000;  //15 seconds
	public static final int MAX_SIMULTANEOUS_CONNECTIONS_PER_HOST = 10;

	
	//TODO rbeede Add support for Content-Encoding: gzip
	//TODO rbeede Add support for Accept-Ranges: bytes, have threshold so if content-length < xxx does not do range
	//TODO rbeede Add support uploading binary content
	//TODO rbeede Add support for proxy server login credentials
	
	/**
	 * @param username Username to use for authentication.  Use null for no authentication.
	 * @param password Password to use for authentication.  Use null for no password.
	 * @param proxySelector ProxySelector that chooses appropriate proxy settings.  If null uses {@link ProxySelector#getDefault()}.
	 * @param progressListener Optional listener that is called with progress on downloads.  May be null.
	 */
	public Client(String username, String password, ProxySelector proxySelector,
			ProgressListener progressListener) {
		super();
		this.username = username;
		this.password = password;
		
		if(null != proxySelector) {
			this.proxySelector = proxySelector;
		} else {
			this.proxySelector = ProxySelector.getDefault();
		}
		
		this.progressListener = progressListener;
		
		
		//Setup the HTTP Client library this library uses
		this.httpClient = new HttpClient(new MultiThreadedHttpConnectionManager());
		
		//The number of connections / host (keeps things sane)
		this.httpClient.getHttpConnectionManager().getParams().setDefaultMaxConnectionsPerHost(MAX_SIMULTANEOUS_CONNECTIONS_PER_HOST);
		
		//Authentication credentials
		if(null != this.username) {
			this.httpClient.getState().setCredentials(AuthScope.ANY, new UsernamePasswordCredentials(this.username, this.password));
		}
		
		//Send authentication credentials before server asks for them (saves time)
		this.httpClient.getParams().setAuthenticationPreemptive(true);

		//Timeouts
		this.httpClient.getHttpConnectionManager().getParams().setConnectionTimeout(TIMEOUT_MS);
		this.httpClient.getHttpConnectionManager().getParams().setSoTimeout(TIMEOUT_MS);
	}


	@Override
	protected void finalize() throws Throwable {
		try {
			this.httpClient.getHttpConnectionManager().closeIdleConnections(0);
			this.httpClient.getState().clear();
		} finally {
			super.finalize();
		}
	}
	
	
	/**
	 * Convenience method that calls {@link #download(URI[], OutputStream[], PostParameter[])} with the necessary
	 * {@link FileOutputStream} for the given files.
	 * 
	 * <p style="font-weight: bold;">Notes:</p>
	 * 
	 * <ul>
	 * 	<li>Files that successfully download will have their last modified time stamp updated if the information
	 * 		is available from the remote server.
	 * 	</li>
	 * 	<li>The array of DownloadResult returned will correspond to the items in uris at the time the method was
	 * 		called.  If the array of uris was modified by another thread the modification will be ignored in the
	 * 		returned result.
	 * 	</li>
	 * 	<li>Underlying values in each parameter array are immutable objects and thus thread safe
	 * 	</li>
	 * 	<li>Blocks
	 * 	</li>
	 * 	<li>You must check the returned DownloadResult for errors.  If the server returned an error any content it
	 * 		provided will be written to the given destination.  Therefore the content may be an error message instead
	 * 		of what you think.
	 * 	</li>
	 * 	<li>For each File in files any parent directories will be created as needed.
	 * 	</li>
	 * </ul>
	 * 
	 * @param uris URIs to download in parallel
	 * @param files Corresponding target File for uris
	 * @param postData If null then HTTP GET is used.  If non-null then each request is done with HTTP POST with
	 * 		all given data in the array.  Array does not correspond to uris or files.
	 * @return Array of DownloadResult with the result of the download corresponding to each item in uris
	 * @throws NullPointerException If uris or files was null
	 */
	public DownloadResult[] download(URI[] uris, File[] files, PostParameter[] postData) {
		if(null == uris) {
			throw new NullPointerException("uris was null");
		} else if(null == files) {
			throw new NullPointerException("files was null");
		}
		if(uris.length == 0) {
			return new DownloadResult[0];
		} else if(uris.length != files.length) {
			throw new IllegalArgumentException("uris.length of " + uris.length + " is not equal to files.length of " + files.length);
		}
		
		//For thread safety make a copy of the passed in arrays (prevents external sorting from messing things up)
		uris = java.util.Arrays.copyOf(uris, uris.length);
		files = java.util.Arrays.copyOf(files, files.length);
		if(null != postData) {
			postData = java.util.Arrays.copyOf(postData, postData.length);
		}
		
		
		//Temporary holder for results
		final DownloadResult[] tempResults;
		
		final OutputStream[] ostreams = new OutputStream[files.length];
		for(int i = 0; i < ostreams.length; i++) {
			//Create any parent folders, if necessary
			if(null != files[i] && null != files[i].getParentFile() && !files[i].getParentFile().exists()) {
				files[i].getParentFile().mkdirs();  //may fail, error later on if so
			}
			
			try {
				ostreams[i] = new FileOutputStream(files[i]);
			} catch (FileNotFoundException e) {
				log.trace(e,e);
				
				ostreams[i] = null;  //signals failed to setup file
			}
		}
		
		
		//Call method to do the actual download
		tempResults = download(uris, ostreams, postData);
		
		
		//We need to convert tempResults to the type that has a reference to the original File[] array
		final DownloadResult[] finalResults = new DownloadResult[tempResults.length];
		for(int i = 0; i < tempResults.length; i++) {
			finalResults[i] = new DownloadResult(
					tempResults[i].getURI(),
					tempResults[i].getServerResponseCode(),
					tempResults[i].getServerResponseString(),
					tempResults[i].getRemoteContentLength(),
					tempResults[i].getTotalBytesRead(),
					tempResults[i].getLastModified(),
					tempResults[i].getError(),
					tempResults[i].getStartTimeEpochMillis(),
					tempResults[i].getEndTimeEpochMillis(),
					files[i]
					);
			
			//Modify the time stamp of the file if available
			if(finalResults[i].getLastModified() >= 0 && null == finalResults[i].getError()) {
				files[i].setLastModified(finalResults[i].getLastModified());
			}
		}
		
		return finalResults;
	}


	/**
	 * <p style="font-weight: bold;">Note:</p>
	 * See {@link #download(URI[], File[], PostParameter[])} for detailed notes on behaviors
	 * 
	 * <p style="font-weight: bold;">Note:</p>
	 * Blocks
	 * 
	 * <p style="font-weight: bold;">Note:</p>
	 * You must check the returned DownloadResult for errors.  If the server returned an error any content it
	 * provided will be written to the given destination.  Therefore the content may be an error message instead
	 * of what you think.
	 * 
 	 * @param uris URIs to download in parallel
	 * @param ostreams Destination streams for uris.  OutputStream#close() is always called.
	 * @param postData If null then HTTP GET is used.  If non-null then each request is done with HTTP POST with
	 * 		all given data in the array.  Array does not correspond to uris or streams.  You may pass in a zero
	 * 		length array to force HTTP POST with no parameters
	 * @return Array of DownloadResult with the result of the download corresponding to each item in uris
	 * @throws NullPointerException if uris or osteams is null
	 * @throws IllegalArgumentException if uris.length != ostreams.length
	 * @see #download(URI[], File[], PostParameter[])
	 */
	public DownloadResult[] download(URI[] uris, OutputStream[] ostreams, PostParameter[] postData) {
		if(null == uris) {
			throw new NullPointerException("uris was null");
		} else if(null == ostreams) {
			throw new NullPointerException("ostreams was null");
		}
		if(uris.length == 0) {
			return new DownloadResult[0];
		} else if(uris.length != ostreams.length) {
			throw new IllegalArgumentException("uris.length of " + uris.length + " is not equal to ostreams.length of " + ostreams.length);
		}
		
		//For thread safety make a copy of the passed in arrays (prevents external sorting from messing things up)
		uris = java.util.Arrays.copyOf(uris, uris.length);
		ostreams = java.util.Arrays.copyOf(ostreams, ostreams.length);
		if(null != postData) {
			postData = java.util.Arrays.copyOf(postData, postData.length);
		}

		
		// Prepare storage of results
		final DownloadResult[] downloadResults = new DownloadResult[uris.length];
		
		final Queue<ProcessingThread> downloadQueue = new LinkedBlockingQueue<ProcessingThread>(uris.length);
		
		// Create and start a thread for each URI
		for(int i = 0; i < uris.length; i++) {
			ProcessingThread thread = new ProcessingThread(new DownloadRunnable(this.httpClient, uris[i], ostreams[i], postData, this.proxySelector, this.progressListener), i);
			
			thread.start();
			
			downloadQueue.offer(thread);
		}
		
		// Track the progress and block until all are done (or have errored out)
		while(!downloadQueue.isEmpty()) {
			final ProcessingThread headThread = downloadQueue.poll();
			
			if(!Thread.State.TERMINATED.equals(headThread.getState())) {
				//Back to the end of the queue
				downloadQueue.offer(headThread);
			} else {
				//Grab the result and store it
				downloadResults[headThread.getIndex()] = headThread.getDownloadRunnable().getDownloadResult();
				
				log.trace(uris[headThread.getIndex()] + " has completed");
				
				//Leave thread off queue so queue shrinks
			}
		}
		
		
		// Return the results
		return downloadResults;
	}
	

	/**
	 * Convenience method
	 * 
	 * @see #download(URI[], File[], PostParameter[])
	 */
	public DownloadResult[] download(URI uri, File file, PostParameter[] postData) {
		return download(new URI[]{uri},new File[]{file}, postData);
	}

	/**
	 * Convenience method
	 * 
	 * @see #download(URI[], OutputStream[], PostParameter[])
	 */
	public DownloadResult[] download(URI uri, OutputStream ostream, PostParameter[] postData) {
		return download(new URI[]{uri},new OutputStream[]{ostream}, postData);
	}
}