package net.sourceforge.chopchophttpclient;

import java.io.File;
import java.io.OutputStream;
import java.net.URI;

/**
 * Immutable class that contains the result of a download.
 * 
 * @author rbeede
 * 
 */
public class DownloadResult {
	private final URI uri; // Immutable according to Java API
	
	private final String serverResponseString;
	private final int serverResponseCode;

	/**
	 * @see #getRemoteContentLength()
	 */
	private final long remoteContentLength;

	/**
	 * @see #getTotalBytesRead()
	 */
	private final long totalBytesRead;

	private final Throwable error;

	private final long lastModified;

	// Content may be written to only one of these
	private final OutputStream outputStreamDestination;
	private final File fileDestination;  //also gets lastModified stamped automatically

	
	private final long startTimeEpochMillis;  //Start time in epoch milliseconds (not seconds)
	private final long endTimeEpochMillis;  //Start time in epoch milliseconds (not seconds)

	
	public DownloadResult(final URI uri, final int serverResponseCode, final String serverResponseString,
			final long remoteContentLength, final long totalBytesRead, final long lastModified, final Throwable error,
			final long startTimeEpochMillis, final long endTimeEpochMillis,
			final OutputStream outputStreamDestination) {
		super();
		
		this.uri = uri;
		this.serverResponseCode = serverResponseCode;
		this.serverResponseString = serverResponseString;
		this.remoteContentLength = remoteContentLength;
		this.totalBytesRead = totalBytesRead;
		this.lastModified = lastModified;
		this.error = error;
		
		this.outputStreamDestination = outputStreamDestination;
		this.fileDestination = null;
		
		this.startTimeEpochMillis = startTimeEpochMillis;
		this.endTimeEpochMillis = endTimeEpochMillis;
	}

	protected DownloadResult(final URI uri, final int serverResponseCode, final String serverResponseString,
			final long remoteContentLength, final long totalBytesRead, final long lastModified, final Throwable error,
			final long startTimeEpochMillis, final long endTimeEpochMillis,
			final File fileDestination) {
		super();
		
		this.uri = uri;
		this.serverResponseCode = serverResponseCode;
		this.serverResponseString = serverResponseString;
		this.remoteContentLength = remoteContentLength;
		this.totalBytesRead = totalBytesRead;
		this.lastModified = lastModified;
		this.error = error;
		
		this.outputStreamDestination = null;
		this.fileDestination = fileDestination;

		this.startTimeEpochMillis = startTimeEpochMillis;
		this.endTimeEpochMillis = endTimeEpochMillis;
	}
	

	public String getServerResponseString() {
		return serverResponseString;
	}

	public int getServerResponseCode() {
		return serverResponseCode;
	}

	/**
	 * <p style="font-weight: bold;">Note:</p>
	 * Even if this returns null it is possible the remote server returned an error in its response.  Use
	 * {@link #getServerResponseCode()} to check for errors from the remote server.
	 * 
	 * @return Error that caused download to completely fail (no response available).  null if no critical error occurred
	 */
	public Throwable getError() {
		return error;
	}

	public long getLastModified() {
		return lastModified;
	}

	/**
	 * @return Handle to OutputStream given for storing content.  Will be null if download was to {@link File} instead.
	 */
	public OutputStream getOutputStreamDestination() {
		return outputStreamDestination;
	}

	/**
	 * @return Handle to File given for storing content.  Will be null if download was to {@link OutputStream} instead.
	 */
	public File getFileDestination() {
		return fileDestination;
	}

	public URI getURI() {
		return uri;
	}

	/**
	 * Content-Length as reported by the remote server.
	 * 
	 * <p>
	 * This could be different from the actual download content bytes read if the response was compressed or if the
	 * remote server did not give the content length.
	 * </p>
	 * 
	 * @return -1 if content length could not be determined.  Length from remote server otherwise.
	 */
	public long getRemoteContentLength() {
		return remoteContentLength;
	}

	/**
	 * Amount of bytes successfully read
	 * 
	 * <p>
	 * Note that this amount could be greater than the content length reported from {@link #getRemoteContentLength()}
	 * if the response was compressed by the server and decompressed by the client.
	 * </p>
	 */
	public long getTotalBytesRead() {
		return totalBytesRead;
	}
	
	
	public long getStartTimeEpochMillis() {
		return this.startTimeEpochMillis;
	}
	
	public long getEndTimeEpochMillis() {
		return this.endTimeEpochMillis;
	}
	
	public long getDownloadTimeSeconds() {
		long diff = (this.endTimeEpochMillis - this.startTimeEpochMillis) / 1000;
		
		if(diff < 0) {
			diff = 0;
		}
		
		return diff;
	}
}